import html2text
import pandas as pd

from scrapy_llm.types import CombinedLLMOutput, T


def process_html(
    html: str,
    ignore_links: bool = True,
    ignore_images: bool = True,
) -> str:
    """
    Process HTML to plain text using html2text.

    Args:
        html (str): The HTML to process.
        ignore_links (bool): Whether to ignore links. Defaults to True.
        ignore_images (bool): Whether to ignore images. Defaults to True.

    Returns:
        str: The processed plain text.
    """
    cleaner = html2text.HTML2Text()
    cleaner.ignore_links = ignore_links
    cleaner.ignore_images = ignore_images
    clean_content = cleaner.handle(html)
    return " ".join(clean_content)


def flatten_dict(items: T) -> CombinedLLMOutput:
    """
    Flatten nested dicts in each item to attributes.

    Example:
    >>> items = [
        {
            "name": "John",
            "address": {"city": "New York", "state": "NY"},
            "contact": {"email": "test@test.com, "phone": "1234567890"}
        }
    ]

    >>> flatten_items(items) -> [
        {
            "name": "John",
            "city": "New York",
            "state": "NY",
            "email": "test@test.com",
            "phone": "1234567890"
        }
    ]
    """

    nested_keys, flat_items = set(), []

    # find all nested keys
    for item in items:
        for key, value in item.items():
            if isinstance(value, dict):
                nested_keys.add(key)

    # avoid messing up the original list
    if not nested_keys:
        return items.pop()

    for item in items:
        item_df = pd.DataFrame([item])
        for key in nested_keys:
            if key not in item_df.columns:
                continue
            item_df = pd.concat(
                [item_df, pd.DataFrame(item_df[key].tolist())], axis=1
            ).drop(key, axis=1)

            flat_items.append(item_df.to_dict(orient="records").pop())

    # single item
    if len(flat_items) == 1:
        return flat_items.pop()

    return flat_items



from typing import Generic, Iterable, List, Type, TypeVar

import instructor
from litellm import completion
from scrapy import Request, signals
from scrapy.crawler import Crawler
from scrapy.http.response import Response
from scrapy.spiders import Spider

from scrapy_llm.config import LlmExtractorConfig
from scrapy_llm.types import LLMOutput, T
from scrapy_llm.utils import flatten_dict, process_html

LLMExtractor = TypeVar("LLMExtractor", bound="LlmExtractorMiddleware")


class LlmExtractorMiddleware(Generic[T]):
    def __init__(
        self,
        crawler: Crawler,
        config: LlmExtractorConfig,
    ):
        self.crawler = crawler
        self.config = config

    @classmethod
    def from_crawler(cls: Type[LLMExtractor], crawler: Crawler) -> LLMExtractor:
        config = LlmExtractorConfig.from_crawler(crawler)
        instance = cls(crawler, config)
        crawler.signals.connect(instance.spider_opened, signal=signals.spider_opened)
        return instance

    def process_response(
        self, request: Request, response: Response, spider: Spider
    ) -> Response:
        extracted_data = self.extract_item_data(
            response.text, self.config.response_model
        )

        if self.config.unwrap_nested:
            extracted_data = flatten_dict(extracted_data)

        request.meta["llm_extracted_data"] = extracted_data

        return response

    def spider_opened(self, spider: Spider) -> None:
        spider.logger.info("Spider opened: %s" % spider.name)

    def extract_item_data(
        self,
        raw_html: str,
        response_model: Type[T],
    ) -> List[LLMOutput]:
        """Extract dorm data from the given HTML text using an LLM model."""

        cl = instructor.from_litellm(completion)

        resp: List[response_model] = cl.chat.completions.create(
            model=self.config.llm_model,
            api_base=self.config.llm_api_base,
            custom_llm_provider="openai",
            temperature=self.config.llm_temperature,
            messages=[
                {
                    "role": "system",
                    "content": self.config.llm_system_message,
                },
                {"role": "user", "content": process_html(raw_html)},
            ],
            # always return an iterable of the given type to simplify nested type cascading.
            response_model=Iterable[response_model],
        )

        items = [item.model_dump() for item in resp]

        return items
